/* -*- C++ -*-
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2002 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

/*
 * 4758_sym_crypto.cc
 * alex iliev, apr 2003
 * implementations for symmetric crypto provider classes using 4758 crypto
 * interface
 */

#include <string>

#include <string.h>
#include <errno.h>

#include <sys/types.h>		// for open(2)
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>

#include <leeds/crypto/crypto.h>

#include "4758_sym_crypto.h"


using namespace std;


SymCrypt4758::SymCrypt4758 () throw (crypto_exception)
    : SymCryptProvider (8, 8)
{
    _fd = open (DES_DEVICEFILE, O_WRONLY);
    if (_fd <= 0) {
	throw crypto_exception
	    (string ("Failed to open 4758 DES device file: ") +
	     strerror(errno));
    }
    
    memset (&_req, 0, sizeof(_req));

    _req.operation = SCC_TDES;
    _req.mode = CBC;
    _req.command = MSGPART_ONLY;
    _req.padding = NOPADDING;	// FIXME: the 4758 code must be screwed because
				// this produces padding (which i want)

    // the other parameters are to be set when doing an operation
}



SymCrypt4758::~SymCrypt4758() {
    close (_fd);
}



void SymCrypt4758::symcrypto_op (const ByteBuffer& input,
				 const ByteBuffer& key,
				 const ByteBuffer& iv,
				 ByteBuffer & out,
				 SymCryptProvider::OpType optype)
    throw (crypto_exception)
{
    _req.encdec = (optype == SymCryptProvider::ENCRYPT ? ::ENCRYPT : ::DECRYPT);

    memcpy (_req.keys, key.data(), key.len()); // this treats the three 8-byte
					       // key fields as one, which
					       // should be fine in C++
    memcpy (_req.IV, iv.data(), iv.len());

    _req.inptr = input.data();
    _req.inLen = input.len();

    _req.outptr = out.data();
    _req.outLen = out.len();

    ssize_t written = write (_fd, &_req, sizeof(_req));
    if (written < 0) {
	throw crypto_exception ("4758 TDES " +
				SymCryptProvider::get_op_name (optype) +
				" failed: " + strerror (errno));
    }

    out.len() = written;
}





//
// class SHA1HMAC_4758
//


SHA1HMAC_4758::SHA1HMAC_4758 () throw (crypto_exception)
    : MacProvider (HMAC_SHA1_MACSIZE)
{
    _fd = open (HMACSHA1_DEVICEFILE, O_RDWR);
    if (_fd <= 0) {
	throw crypto_exception
	    (string ("Failed to open 4758 HMAC SHA1 device file: ") +
	     strerror(errno));
    }

    memset (&_req, 0, sizeof(_req));

    _req.operation = SCC_HMACSHA1;
    _req.command = MSGPART_ONLY;
}


SHA1HMAC_4758::~SHA1HMAC_4758 () {
    close (_fd);
}


void SHA1HMAC_4758::genmac (const ByteBuffer& text, const ByteBuffer& key,
			    ByteBuffer & out)
    throw (crypto_exception)
{
    
    _req.data = text.data();
    _req.data_len = text.len();

    memcpy (_req.key, key.data(), key.len());
    _req.key_len = key.len();

    _req.hash = out.data();
    _req.hash_len = out.len();

    if ( write (_fd, &_req, sizeof(_req)) < 0 ) {
	throw crypto_exception (string("4758 HMAC SHA1 failed: ") +
				strerror (errno));
    }

}
