/*
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2002 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

/*
 * pirserver.cc
 * alex iliev, jan 2003
 * the card-side code to handle PIR requests from outside
 */

#include <string>
#include <vector>
#include <new>
#include <stdexcept>

#include <sstream>

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>

#include <unistd.h>


#include <common/utils.h>
#include <common/sym_crypto.h>
#include <common/consts.h>
#include <common/exceptions.h>
#include <common/socket-class.h>
#include <common/countptr.hpp>

#include <common/openssl_crypto.h>


#include "lib.h"
#include "pirsession.h"
#include "consts.h"
#include "4758_sym_crypto.h"


using namespace std;


// do one retrieval session using one shuffled DB
// may add newly announced DBs to dbdirs
void do_session (PIRSession & session, list<string> & io_dbdirs,
		 DatagramFDSocket & search_sock,
		 DatagramFDSocket & control_sock,
		 SocketSelector & sel)
    throw (comm_exception);


static void handle_control (DatagramSocket & sock, list<string> & db_dirs)
    throw (comm_exception);

static void handle_search (DatagramSocket & sock, PIRSession & session)
    throw (comm_exception);




// HACK: ugh, global
char * g_program;

// these two stolen from sccping.c from IBM
static void
error(char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    fprintf(stderr, "%s: Error: ", g_program);
    vfprintf(stderr, fmt, ap);
    fputc('\n', stderr);
    va_end(ap);
    exit(EXIT_FAILURE);
}

static void
warning(char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    fprintf(stderr, "%s: Warning: ", g_program);
    vfprintf(stderr, fmt, ap);
    fputc('\n', stderr);
    va_end(ap);
}



static void
usage(const char* progname)
{
    fprintf(stderr,
	    "Usage: %s\n"
	    "[-p <search port>]\n"
	    "[-d <root directory>]\n"
	    "[-i <shuffled DB dir> (can have many -i's, one per DB)]\n"
	    "[-c (means to use the card's crypto hardware)]\n",
	    progname);
    exit(EXIT_FAILURE);
}



void out_of_memory () {
    cerr << "Out of memory!" << endl;
}


static void exception_exit (const std::exception& ex, const string& msg) {
    	cerr << msg << endl
	     << ex.what() << endl
	     << "Exiting ..." << endl;
	exit (EXIT_FAILURE);
}



int main (int argc, char *argv[]) {

    int port = PIR_SEARCH_PORT;

    string dir = ROOTDIR;

    bool use_card_crypto = false;

    list<string> db_dirs;	// the queue of shuffled databases (ie their
				// directry names) we have currently.
				// enqueue on the back, dequeue from front


    // from sccping.c
    int opt;
    while ((opt = getopt(argc, argv, "p:d:i:c")) != EOF) {
	switch (opt) {

        case 'p':		// port
	    port = strtol(optarg, NULL, 0);
	    if (port == 0)
		error("invalid port");
	    break;

	case 'd':		//directory for keys etc
	    dir = optarg;
	    break;

	case 'c':
	    use_card_crypto = true;
	    break;

	case 'i':
	    db_dirs.push_back (optarg);
	    break;

	default:
	    usage(argv[0]); 
	}
    }
    
    // read in the initial databases, which should be at the end of
    // the argv array now
    while (optind < argc) {
	clog << "Initial DB " << argv[optind] << endl;
	db_dirs.push_back (argv[optind++]);
    }


    g_program = argv[0];


    set_new_handler (out_of_memory);

    
    //
    // initialize everything
    //

    
    // setup the sockets
    SCCDatagramSocket search_sock;
    SCCDatagramSocket control_sock;
    try {
	search_sock.bind (SCCSocketAddress (port));
	control_sock.bind (SCCSocketAddress (PIR_CONTROL_PORT));
    }
    catch (const comm_exception& ex) {
	exception_exit (ex, "Error setting up server sockets");
    }
    

    // buffers for our keys
    ByteBuffer keybuf, mackeybuf;
    try {
	// TODO: for now we're fetching the keys off the host, this needs to be
	// implemented for real!
	loadkeys (keybuf,    dir + DIRSEP + KEYFILE,
		  mackeybuf, dir + DIRSEP + MACKEYFILE);
    }
    catch (const ios_base::failure & ex) {
	exception_exit (ex, "Error fetching keys");
    }
    

    //
    // initialize the crypto providers
    //
    auto_ptr<SymCryptProvider> symop;
    auto_ptr<MacProvider> macop;
    try {
	if (use_card_crypto) {
	    symop.reset (new SymCrypt4758 ());
	    macop.reset (new SHA1HMAC_4758());
	}
	else {
	    symop.reset (new OSSLSymCrypto());
	    macop.reset (new OSSL_HMAC());
	}
    }
    catch (const crypto_exception& ex) {
	exception_exit (ex, "Error making crypto providers");
    }
	
    

    // if no starting DB's, just wait on the control socket until one comes
    // along
    if (db_dirs.empty()) {
	clog << "Waiting for first shuffled DB..." << endl;
	try {
	    handle_control (control_sock, db_dirs);
	}
	catch (const comm_exception& ex) {
	    exception_exit (ex, "Getting initial shuffled database name");
	}
    }


    clog << "Starting to listen for requests" << endl;
	
    // make the selector/multiplexor object
    SocketSelector sel;
    try {
	sel.add (search_sock);
	sel.add (control_sock);
    }
    catch (const comm_exception& ex) {
	exception_exit (ex, "Initializing socket multiplexing");
    }
    


    try {

	// loop over the sessions
	while (true) {
	    
	    // make the session object
	    string dbdir = db_dirs.front();
	    db_dirs.pop_front();
	    PIRSession session (dbdir,
				*symop, *macop,
				keybuf, mackeybuf, BUCKETSIZE);

	    do_session (session, db_dirs,
			search_sock, control_sock, sel);

	    clog << "Done with session, moving onto next" << endl;

	    // TODO: done with this session/DB, so tell host to dispose of the
	    // DB

	}
	
    }
    catch (const std::exception& ex) {
	// FIXME: this is a little too brittle, should handle errors more
	// carefully
	exception_exit (ex, "Handling retrieval sessions");
    }

}





// do_session can modify dbdirs to add any newly arrived db
// directories
void do_session (PIRSession & session, list<string> & dbdirs,
		 DatagramFDSocket & search_sock,
		 DatagramFDSocket & control_sock,
		 SocketSelector & sel)
    throw (comm_exception)
{
    // loop over the requests in this session
    while (!session.had_enough() || dbdirs.empty()) {
		
	sel.select (-1);

//	clog << "Select returned!" << endl;

	if (sel.hasData (control_sock)) {
	    handle_control (control_sock, dbdirs);
	}

	if (sel.hasData (search_sock)) {
	    // got a search request
	    handle_search (search_sock, session);
	}
		
    }
}


		 

void handle_control (DatagramSocket & sock, list<string> & db_dirs)
    throw (comm_exception)
{
    counted_ptr<SocketAddress> peer_addr;
	
    // FIXME: this control message has to be better authenticated!
    ByteBuffer msg = sock.recv(peer_addr);

    string db_dirname (msg.cdata(), msg.len());

    clog << "Got a new shuffled DB " << db_dirname << "!" << endl;

    db_dirs.push_back (db_dirname);
}

    


void handle_search (DatagramSocket & sock, PIRSession & session)
    throw (comm_exception)
{

//    clog << "About to receive a search request!" << endl;
    
    counted_ptr<SocketAddress> peer_addr;
    
    /* get request */
    ByteBuffer req = sock.recv(peer_addr);

    string name (req.cdata(), req.len());

    Record rec;
    clog << "Retrieving name " << name << endl;
    rec = session.find_record (name);
    clog << "Done" << endl;

    // FIXME: non-standard serialization procedure here!
    // it's partially because for now the host pirsearch program just
    // displays this string
    ostringstream os;
    os << rec;

    try {
	sock.send ( ByteBuffer(os.str()), *peer_addr );
    }
    catch (const comm_exception& ex) {
	cerr << "Failed to send search response: " << ex.what() << endl;
    }
}
